import 'dart:async';
import 'dart:typed_data';

import 'package:crypto/crypto.dart';
import 'package:cryptography/cryptography.dart' as cryptography;
import 'package:ed25519_edwards/ed25519_edwards.dart' as ed25519;
import 'package:x25519/x25519.dart' as x25519;
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/utils.dart';

Future<String> encrypt(B64 key, List<int> blob) async {
  var box = await cryptography.Xchacha20.poly1305Aead()
      .encrypt(blob, secretKey: cryptography.SecretKey(key.decode().toList()));
  var text = B64.encode(box.cipherText).toString();
  var mac = B64.encode(box.mac.bytes).toString();
  var nonce = B64.encode(box.nonce).toString();

  return "$nonce.$mac.$text";
}

Future<B64> deriveKey(B64 mySK, B64 otherSidePK) async {
  return B64.encode(x25519.X25519(mySK.decode().sublist(0, 32).toList(),
      otherSidePK.decode().sublist(32).toList()));
}

B64 randKey() {
  return B64.encode(randBytes(32).toList());
}

Future<List<int>> decrypt(B64 key, String box) {
  var parts = box.split(".");
  if (parts.length != 3) {
    throw ExInvalidCipherText;
  }
  var nonce = B64(parts[0]).decode();
  var mac = B64(parts[1]).decode();
  var text = B64(parts[2]).decode();
  return cryptography.Xchacha20.poly1305Aead().decrypt(
      cryptography.SecretBox(text.toList(),
          nonce: nonce.toList(), mac: cryptography.Mac(mac.toList())),
      secretKey: cryptography.SecretKey(key.decode().toList()));
}

B64 digest(List<int> blob) {
  return B64.encode(sha256.convert(blob).bytes);
}

B64 genSK() {
  var pair = x25519.generateKeyPair();
  var sk = ed25519.newKeyFromSeed(Uint8List.fromList(pair.privateKey));
  var list = List<int>.filled(32 + 32 + 32, 0);
  list.setRange(0, 64, sk.bytes);
  list.setRange(64, 64 + 32, pair.publicKey);
  return B64.encode(list);
}

B64 derivePK(B64 secretKey) {
  var sk = secretKey.decode();
  if (sk.length < ed25519.PrivateKeySize) {
    throw ExInvalidSK;
  }
  return B64.encode(sk.sublist(32).toList());
}

class ExInvalidSK implements Exception {
  String errMsg() => "invalid secret key";
}

class ExInvalidCipherText implements Exception {
  String errMsg() => "invalid cipher text";
}

class ExInvalidSign implements Exception {
  String errMsg() => "invalid signature";
}

B64 sign(B64 sk, Uint8List data) {
  var key = ed25519.PrivateKey(sk.decode().sublist(0, 64).toList());
  var sig = ed25519.sign(key, data);
  return B64.encode(sig.toList());
}

void verify(Uint8List data, B64 signature, B64 pk) {
  if (!ed25519.verify(ed25519.PublicKey(pk.decode().sublist(0, 32)), data,
      signature.decode())) {
    throw ExInvalidSign();
  }
}
